package ru.dbotthepony.mc.otm.worldgen

import com.github.benmanes.caffeine.cache.Cache
import com.github.benmanes.caffeine.cache.Caffeine
import com.github.benmanes.caffeine.cache.Scheduler
import com.mojang.datafixers.util.Either
import com.mojang.datafixers.util.Pair
import com.mojang.serialization.Codec
import com.mojang.serialization.DataResult
import com.mojang.serialization.DynamicOps
import com.mojang.serialization.codecs.RecordCodecBuilder
import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import it.unimi.dsi.fastutil.objects.ObjectRBTreeSet
import net.minecraft.Util
import net.minecraft.core.BlockPos
import net.minecraft.core.Holder
import net.minecraft.core.SectionPos
import net.minecraft.world.level.ChunkPos
import net.minecraft.world.level.WorldGenLevel
import net.minecraft.world.level.levelgen.feature.ConfiguredFeature
import net.minecraft.world.level.levelgen.feature.Feature
import net.minecraft.world.level.levelgen.feature.FeaturePlaceContext
import net.minecraft.world.level.levelgen.feature.configurations.FeatureConfiguration
import net.minecraft.world.level.levelgen.placement.PlacedFeature
import net.minecraft.world.level.levelgen.placement.PlacementModifier
import ru.dbotthepony.kommons.util.XXHash64
import ru.dbotthepony.mc.otm.core.collect.BlockPosHashStrategy
import ru.dbotthepony.mc.otm.core.util.GJRAND64RandomSource
import ru.dbotthepony.mc.otm.data.codec.minRange
import ru.dbotthepony.mc.otm.worldgen.feature.EnhancedFeature
import ru.dbotthepony.mc.otm.worldgen.placement.EnhancedPlacementModifier
import java.io.DataOutputStream
import java.time.Duration
import java.util.IdentityHashMap
import kotlin.math.sqrt

private object NodeCodec : Codec<EnhancedPlacedFeature.Node> {
	private val list = Codec.list(this)

	override fun <T : Any?> encode(input: EnhancedPlacedFeature.Node, ops: DynamicOps<T>, prefix: T): DataResult<T> {
		val result = HashMap<T, T>()

		if (input.children.isNotEmpty()) {
			val re = list.encodeStart(ops, input.children)

			if (re.isError)
				return re

			result[ops.createString("children")] = re.resultOrPartial().get()
		}

		if (input.contents.left().isPresent) {
			val re = EnhancedPlacementModifier.CODEC.encodeStart(ops, input.contents.left().get())

			if (re.isError)
				return DataResult.error { "Failed to serialize placement modifier: ${re.error().get().message()}" }

			result[ops.createString("modifier")] = re.resultOrPartial().get()
		} else {
			val re = EnhancedFeature.CODEC.encodeStart(ops, input.contents.right().get())

			if (re.isError)
				return DataResult.error { "Failed to serialize feature: ${re.error().get().message()}" }

			result[ops.createString("feature")] = re.resultOrPartial().get()
		}

		return ops.mergeToMap(prefix, result)
	}

	override fun <T : Any?> decode(ops: DynamicOps<T>, input: T): DataResult<Pair<EnhancedPlacedFeature.Node, T>> {
		return ops.getMap(input).flatMap {
			val children = it.get("children")
			val modifier = it.get("modifier")
			val feature = it.get("feature")

			if (modifier != null && feature != null) {
				return@flatMap DataResult.error { "Enhanced feature placement node should have either modifier or feature, but both are present" }
			} else if (modifier == null && feature == null) {
				return@flatMap DataResult.error { "Enhanced feature placement node contains no modifier and no feature" }
			}

			val deserializeChildren: List<EnhancedPlacedFeature.Node>

			if (children != null) {
				val result = list.decode(ops, children)

				if (result.isError)
					return@flatMap DataResult.error { result.error().get().message() }

				deserializeChildren = result.map { it.first }.getOrThrow()
			} else {
				deserializeChildren = emptyList()
			}

			if (modifier != null) {
				return@flatMap EnhancedPlacementModifier.CODEC.decode(ops, modifier).map { Pair(EnhancedPlacedFeature.Node(deserializeChildren, Either.left(it.first)), ops.empty()) }
			} else {
				return@flatMap EnhancedFeature.CODEC.decode(ops, feature).map { Pair(EnhancedPlacedFeature.Node(deserializeChildren, Either.right(it.first)), ops.empty()) }
			}
		}
	}
}

object EnhancedPlacedFeature : Feature<EnhancedPlacedFeature.Config>(
	RecordCodecBuilder.create {
		it.group(
			Codec.INT.minRange(0).fieldOf("chunk_scan_range").forGetter(Config::chunkScanRange),
			Codec.LONG.fieldOf("seed_mix").forGetter(Config::seedMix),
			NodeCodec.fieldOf("root").forGetter(Config::root),
		).apply(it, ::Config)
	}
) {
	class Node(val children: List<Node>, val contents: Either<EnhancedPlacementModifier, Holder<EnhancedFeature.Configured<*, *>>>){
		fun evaluate(context: EnhancedPlacementContext) {
			evaluate(context, listOf(BlockPos(context.origin.minBlockX, 0, context.origin.minBlockZ)))
		}

		private fun evaluate(context: EnhancedPlacementContext, positions: List<BlockPos>) {
			val actualPositions = if (contents.left().isPresent) {
				contents.left().get().evaluate(context, positions)
			} else {
				positions
			}

			if (contents.right().isPresent) {
				context.place(actualPositions, contents.right().get().value())
			}

			children.forEach {
				it.evaluate(context.push(), actualPositions)
			}
		}
	}

	class Config(
		val chunkScanRange: Int,
		val seedMix: Long,
		val root: Node,
	) : FeatureConfiguration {
		private class GeneratedChunk : EnhancedPlacementContext.Placer {
			private data class Placement(val context: EnhancedPlacementContext, val positions: Set<BlockPos>, val feature: EnhancedFeature.Configured<*, *>)

			// TODO: extremely inefficient
			private val placed = ArrayList<Placement>()

			override fun place(context: EnhancedPlacementContext, positions: List<BlockPos>, feature: EnhancedFeature.Configured<*, *>) {
				placed.add(Placement(context, ObjectRBTreeSet(BlockPosHashStrategy).also { it.addAll(positions) }, feature))
			}

			fun place(context: FeaturePlaceContext<*>): Boolean {
				var any = false
				val pos = ChunkPos(context.origin())

				for ((eContext, positions, feature) in placed) {
					val filtered = positions.filter { SectionPos.blockToSectionCoord(it.x) == pos.x && SectionPos.blockToSectionCoord(it.z) == pos.z }.toTypedArray()

					if (filtered.isNotEmpty()) {
						any = feature.place(eContext, ObjectArraySet.ofUnchecked(*filtered), positions) || any
					}
				}

				return any
			}
		}

		private val level2cache = Caffeine.newBuilder()
			.scheduler(Scheduler.systemScheduler())
			.executor(Util.backgroundExecutor())
			.expireAfterAccess(Duration.ofMinutes(10))
			.weakKeys()
			.build<WorldGenLevel, Cache<ChunkPos, GeneratedChunk>>()

		private fun getCache(level: WorldGenLevel): Cache<ChunkPos, GeneratedChunk> {
			return level2cache.get(level) {
				Caffeine.newBuilder()
					.scheduler(Scheduler.systemScheduler())
					.executor(Util.backgroundExecutor())
					.maximumSize(16384L)
					.expireAfterWrite(Duration.ofMinutes(1))
					.softValues()
					.build()
			}
		}

		private fun evaluate(context: FeaturePlaceContext<*>, chunkPos: ChunkPos): GeneratedChunk {
			val bytes = FastByteArrayOutputStream()
			val dataStream = DataOutputStream(bytes)

			dataStream.writeLong(seedMix)
			dataStream.writeInt(chunkPos.x)
			dataStream.writeInt(chunkPos.z)

			val hash = XXHash64()
			hash.update(bytes.array, 0, bytes.length)

			val random = GJRAND64RandomSource(context.level().seed, hash.digestAsLong())

			val chunk = GeneratedChunk()
			val enhancedContext = EnhancedPlacementContext(context.level(), context.chunkGenerator(), random, chunkPos, chunk)
			root.evaluate(enhancedContext)
			return chunk
		}

		fun place(context: FeaturePlaceContext<*>): Boolean {
			val cache = getCache(context.level())
			val chunkPos = ChunkPos(context.origin())
			val instances = ArrayList<GeneratedChunk>()

			for (x in -chunkScanRange .. chunkScanRange) {
				for (z in -chunkScanRange .. chunkScanRange) {
					// floor, so chunk scan range of 1 will give square instead of diamond
					val radius = sqrt(x.toDouble() * x + z.toDouble() * z).toInt()

					if (radius <= chunkScanRange) {
						val thisPos = ChunkPos(chunkPos.x + x, chunkPos.z + z)
						instances.add(cache.get(thisPos) { evaluate(context, thisPos) })
					}
				}
			}

			var any = false
			instances.forEach { any = it.place(context) }
			return any
		}
	}

	override fun place(context: FeaturePlaceContext<Config>): Boolean {
		return context.config().place(context)
	}

	class Builder {
		private val root: Builder?
		private val children = ArrayList<Builder>()
		private val contents: Either<EnhancedPlacementModifier, Holder<EnhancedFeature.Configured<*, *>>>

		constructor() : this(EnhancedPlacementModifier.Passthrough)

		constructor(root: EnhancedPlacementModifier) {
			this.root = null
			this.contents = Either.left(root)
		}

		constructor(root: PlacementModifier) : this(EnhancedPlacementModifier.Wrapper(root))

		constructor(root: Holder<EnhancedFeature.Configured<*, *>>) {
			this.root = null
			this.contents = Either.right(root)
		}

		constructor(root: EnhancedFeature.Configured<*, *>) {
			this.root = null
			this.contents = Either.right(Holder.direct(root))
		}

		private constructor(parent: Builder, root: PlacementModifier) : this(parent, EnhancedPlacementModifier.Wrapper(root))

		private constructor(parent: Builder, root: EnhancedPlacementModifier) {
			this.root = parent
			this.contents = Either.left(root)
		}

		private constructor(parent: Builder, root: Holder<EnhancedFeature.Configured<*, *>>) {
			this.root = parent
			this.contents = Either.right(root)
		}

		private constructor(parent: Builder, root: EnhancedFeature.Configured<*, *>) {
			this.root = parent
			this.contents = Either.right(Holder.direct(root))
		}

		fun then(action: PlacementModifier): Builder {
			return Builder(root ?: this, action).also(children::add)
		}

		fun then(action: EnhancedPlacementModifier): Builder {
			return Builder(root ?: this, action).also(children::add)
		}

		fun then(action: EnhancedFeature.Configured<*, *>): Builder {
			return Builder(root ?: this, action).also(children::add)
		}

		fun then(action: Holder<EnhancedFeature.Configured<*, *>>): Builder {
			return Builder(root ?: this, action).also(children::add)
		}

		@JvmName("thenVanilla")
		fun then(action: ConfiguredFeature<*, *>): Builder {
			return then(EnhancedFeature.Wrapper.configure(action))
		}

		@JvmName("thenVanilla")
		fun then(action: Holder<ConfiguredFeature<*, *>>): Builder {
			return then(EnhancedFeature.Wrapper.configure(action))
		}

		private fun buildNode(): Node {
			return Node(children.map { it.buildNode() }, contents)
		}

		fun build(
			chunkScanRange: Int,
			seedMix: Long,
		): PlacedFeature {
			if (root != null)
				return root.build(chunkScanRange, seedMix)

			return PlacedFeature(Holder.direct(ConfiguredFeature(EnhancedPlacedFeature, Config(chunkScanRange, seedMix, buildNode()))), listOf())
		}
	}
}
